{-|
Module      : Test.Data.Patch.Class
Description :
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

-}

module Test.Data.Patch.Class where

import Data.Patch.Class
import Data.Validity
import Test.QuickCheck


-- Tests for the Diff class

-- | act . diff = id
prop_act_diff_is_id :: Int -> Int -> Bool
prop_act_diff_is_id d e = act (diff d e :: Replace Int) d == e

-- | diff d d == mempty
prop_act_diff_diff_is_mempty :: Int -> Bool
prop_act_diff_diff_is_mempty d = (diff d d :: Replace Int) == mempty

prop_act_diff_is_act_invert_diff :: Int -> Int -> Bool
prop_act_diff_is_act_invert_diff d e =
  act (diff d e :: Replace Int) d == act (invert (diff e d :: Replace Int)) d

prop_act_diff_composable :: Int -> Int -> Int -> Bool
prop_act_diff_composable a b c =
  act (diff b c <> diff a b :: Replace Int) a == act (diff a c :: Replace Int) a

prop_diff_is_applicable :: Int -> Int -> Bool
prop_diff_is_applicable a b = validationIsValid $ applicable (diff a b :: Replace Int) a

prop_diff_of_two_simple_values :: Int -> Int -> Property
prop_diff_of_two_simple_values a b =
  (a /= b) ==> diff a b == Replace { _old = a, _new = b }

prop_diff_of_same_values :: Int -> Bool
prop_diff_of_same_values a = diff a a == (Keep :: Replace Int)


-- Data.Patch.Class tests

prop_simple_act_works :: Int -> Int -> Bool
prop_simple_act_works o n =
  act (Replace { _old = o, _new = n }) o == n

prop_simple_validation_works :: Int -> Int -> Int -> Bool
prop_simple_validation_works a o n =
  (validationIsValid $ applicable (Replace { _old = o, _new = n }) a) == (a == o)

prop_keep_does_nothing :: Int -> Bool
prop_keep_does_nothing o = act (Keep :: Replace Int) o == o

prop_simple_composable :: Int -> Int -> Int -> Property
prop_simple_composable a b c =
  -- TODO Is this correct? i.e. shouldn't it be @composable p q@?
  (a /= b) && (b /= c) ==> validationIsValid (composable q p)
  where
    p = Replace { _old = a, _new = b }
    q = Replace { _old = b, _new = c }


-- @transformWith@ tests
prop_simple_transform_with_same_old_first_has_priority :: Int -> Int -> Int -> Property
prop_simple_transform_with_same_old_first_has_priority o b c =
  (o /= b) && (o /= c) && (b /= c) ==>
  transformWith ours p q == transformed
  where
    p = Replace { _old = o, _new = b } :: Replace Int
    q = Replace { _old = o, _new = c } :: Replace Int
    transformed = ( Replace { _old = c, _new = b }, Keep )

prop_simple_transform_with_same_old_second_has_priority :: Int -> Int -> Int -> Property
prop_simple_transform_with_same_old_second_has_priority o b c =
  (o /= b) && (o /= c) && (b /= c) ==>
  transformWith theirs p q == transformed
  where
    p = Replace { _old = o, _new = b } :: Replace Int
    q = Replace { _old = o, _new = c } :: Replace Int
    transformed = ( Keep, Replace { _old = b, _new = c } )

prop_simple_transform_with_different_old_first_has_priority :: Int -> Int -> Int -> Int -> Property
prop_simple_transform_with_different_old_first_has_priority o1 o2 b c =
  (o1 /= o2) && (o1 /= b) && (o1 /= c) && (o2 /= b) && (o2 /= c) && (b /= c) ==>
  transformWith ours p q == transformed
  where
    p = Replace { _old = o1, _new = b } :: Replace Int
    q = Replace { _old = o2, _new = c } :: Replace Int
    transformed = ( Replace { _old = c, _new = b }, Keep )

prop_simple_transform_with_different_old_second_has_priority :: Int -> Int -> Int -> Int -> Property
prop_simple_transform_with_different_old_second_has_priority o1 o2 b c =
  (o1 /= o2) && (o1 /= b) && (o1 /= c) && (o2 /= b) && (o2 /= c) && (b /= c) ==>
  transformWith theirs p q == transformed
  where
    p = Replace { _old = o1, _new = b } :: Replace Int
    q = Replace { _old = o2, _new = c } :: Replace Int
    transformed = ( Keep, Replace { _old = b, _new = c } )


-- | @MaybePatch@ tests
prop_simple_maybe_patch_n :: Bool
prop_simple_maybe_patch_n =
  let p = Mod Keep :: MaybePatch Int (Replace Int) in
  act p (Nothing :: Maybe Int) == Nothing

prop_simple_maybe_patch_j :: Int -> Bool
prop_simple_maybe_patch_j a =
  let p = Mod Keep :: MaybePatch Int (Replace Int) in
  act p (Just a) == Just a

prop_simple_maybe_patch_replace_n :: Int -> Int -> Bool
prop_simple_maybe_patch_replace_n o n =
  let p = Mod (Replace { _old = o, _new = n }) :: MaybePatch Int (Replace Int) in
  act p (Nothing :: Maybe Int) == Nothing

-- TODO Is this correct? Shouldn't this validate a == o first?
prop_simple_maybe_patch_replace_j :: Int -> Int -> Int -> Bool
prop_simple_maybe_patch_replace_j a o n =
  let p = Mod (Replace { _old = o, _new = n }) :: MaybePatch Int (Replace Int) in
  act p (Just a) == Just n

prop_simple_maybe_patch_validation_n :: Int -> Int -> Bool
prop_simple_maybe_patch_validation_n o n =
  let p = Mod (Replace { _old = o, _new = n }) :: MaybePatch Int (Replace Int) in
    validationIsValid $ applicable p (Nothing :: Maybe Int)

prop_simple_maybe_patch_validation_j :: Int -> Int -> Int -> Bool
prop_simple_maybe_patch_validation_j a o n =
  let p = Mod (Replace { _old = o, _new = n }) :: MaybePatch Int (Replace Int) in
  (validationIsValid $ applicable p (Just a)) == (a == o)
